/************************************************************
 * $Id: forest.c 599 2007-04-27 15:46:45Z clerger $
 * Copyright (C) 1997, 2003, 2004, 2006, 2007 Eric de la Clergerie
 * ------------------------------------------------------------
 *
 *   Forest --
 *
 * ------------------------------------------------------------
 * Description
 *
 * ------------------------------------------------------------
 */

#include <string.h>
#include "libdyalog.h"
#include "builtins.h"

obj_t seen;
static obj_t indirect;

#define B_TYPE( b )  (backptr_t) (PAIRP( b ) ? CINT( CAR( b )) : CINT( b ))
#define B_CAR( b )   (CAR( CDR( b )))
#define B_CDR( b )   (CDR( CDR( b )))

/**********************************************************************
 *
 **********************************************************************/

static long
get_index( tabobj_t o )
{
  long i = 0;
  obj_t *l = &seen;

  for( ; PAIRP( *l ) && CAR((*l)) != (obj_t) o ; i++, l = &CDR((*l)) );
  
  if (NULLP( *l ))
    *l = MAKE_PAIR( (obj_t) o, BNIL );

  return i;
}

static tabobj_t
starter( tabobj_t o ) 
{
    if (!o)
        return (tabobj_t)0;
    else{
        obj_t backptr = o->backptr;
        if (B_TYPE( backptr ) == f_trace) {
            backptr = B_CDR(backptr);
        }
        switch (B_TYPE( backptr )) {
            case f_init:
            case f_call:
                return (tabobj_t) 0;
                break;
            default:
                return o;
                break;
        }
    }
}

/* EVDLC -- October 5 2006
   The 2 following functions try to remove duplicate derivations in forest.
   It remains to be checked if the process can loop in case of cyclic forest.
   Also the process could possibly made more efficient by flagging reduced
   derivations
*/

static obj_t forest_remove_duplicates( obj_t b);

static Bool
forest_check_duplicate( obj_t b1, obj_t b2 )
{
        // assume b1 and b2 have been already reduced
    if (B_TYPE( b2 ) == f_or) {
        return forest_check_duplicate(b1,B_CAR(b2))
            || forest_check_duplicate(b1,B_CDR(b2));
    } else if (B_TYPE(b1) == f_or) {
        return forest_check_duplicate(B_CAR(b1),b2)
            && forest_check_duplicate(B_CDR(b1),b2);
    } else if (B_TYPE( b2 ) != B_TYPE(b1)) {
        return 0;
    } else {
        switch (B_TYPE( b1 )) {
            case f_init:
            case f_call:
                return 1;
            case f_join:
                return ( starter( (tabobj_t) B_CDR(b1) ) == starter( (tabobj_t) B_CDR(b2) )
                         && starter( (tabobj_t) B_CAR(b1) ) == starter( (tabobj_t) B_CAR(b2) ));
            case f_rand:
            case f_and:
                if ( starter( (tabobj_t) B_CDR(b1) ) != starter( (tabobj_t) B_CDR(b2) ) ) {
                    return 0;
                } else {
                    ((tabobj_t)B_CAR(b1))->backptr = forest_remove_duplicates(((tabobj_t)B_CAR(b1))->backptr);
                    ((tabobj_t)B_CAR(b2))->backptr = forest_remove_duplicates(((tabobj_t)B_CAR(b2))->backptr);
                    return forest_check_duplicate( ((tabobj_t)B_CAR(b1))->backptr,
                                                   ((tabobj_t)B_CAR(b2))->backptr );
                }
            case f_indirect:
                return forest_check_duplicate(B_CAR(b1),B_CAR(b2));
            case f_trace:
                return B_CAR(b1) == B_CAR(b2) && forest_check_duplicate(B_CDR(b1),B_CDR(b2));
            default:
                return 0;
        }
    }
}

static obj_t
forest_remove_duplicates( obj_t b ) 
{
    if (B_TYPE( b ) == f_or) {
            /* b is a disjunction with car simple (not disj)
               remove duplicates from cdr
               and then check if car occur in reduced cdr b2
            */
        obj_t b2 = forest_remove_duplicates( B_CDR(b) );
        if (forest_check_duplicate(B_CAR(b),b2)) {
            return b2;
        } else if (b2 == B_CDR(b)) {
            return b;
        } else {
            return MAKE_PAIR( BINT( (unsigned long) f_or ),
                              MAKE_PAIR( B_CAR(b), b2 ));
        }
    } else if (B_TYPE(b) == f_trace) {
        obj_t b2 = forest_remove_duplicates( B_CDR(b));
        if (b2 == B_CDR(b))
            return b;
        else
            return MAKE_PAIR( BINT( (unsigned long) f_trace),
                              MAKE_PAIR( B_CAR(b), b2));
    } else {
        return b;
    }
}

#define INDIRECT_FOUND 1
#define INDIRECT_VOID 2
#define INDIRECT_CYCLE 4
#define INDIRECT_TEST_CYCLE 8

static Bool forest_void( obj_t, Bool );

static void indirect_check_cycle( obj_t );


static long
find_or_add_indirect( obj_t o, long *info, long init_flag, obj_t *entry)
{
    long i=0;
    obj_t *l =&indirect;
    
    for( ; PAIRP( *l ) && CAR(CAR((*l))) != o ; i++, l = &CDR((*l)) );

//   dyalog_printf("Registering indirect %x %d\n",o,i); /* trace */
        
    if (NULLP( *l )){
        *l = MAKE_PAIR(MAKE_PAIR(o,(obj_t)(init_flag |INDIRECT_TEST_CYCLE)), BNIL );
        * info =  forest_void( B_CAR(o), 1) ? INDIRECT_VOID : 0;
        indirect_check_cycle( B_CAR(o) );
        if ((long)CDR(CAR(*l)) & INDIRECT_CYCLE) {
            *info |= INDIRECT_CYCLE;
        }
        CDR(CAR(*l)) = (obj_t) (init_flag | (*info));
    } else {
        if (((long) CDR(CAR(*l))) & INDIRECT_TEST_CYCLE) {
                // we have found a cycling indirect backptr
            CDR(CAR(*l)) = (obj_t) ((long)  CDR(CAR(*l)) | INDIRECT_CYCLE);
        }
        * info = (long) CDR(CAR(*l));
    }
    * entry = * l;
    return i;
}

static void
indirect_check_cycle( obj_t backptr ) 
{
    if (!backptr) {
        return 1;
    } else {
        switch (B_TYPE( backptr )) {
            case f_or:
                indirect_check_cycle( B_CAR(backptr) );
                indirect_check_cycle(B_CAR(backptr));
                break;
            case f_join:
            case f_rand:
                break;
            case f_and:
                indirect_check_cycle(((tabobj_t)B_CAR(backptr))->backptr);
                break;
            case f_trace:
                indirect_check_cycle( B_CDR(backptr));
                break;
            case f_indirect:
            {
                long info=0;
                obj_t entry=0;
                find_or_add_indirect(backptr,&info,0,&entry);
                break;
            }
            default:
                return 0;
                break;
        }
    }
}


static Bool
forest_void( obj_t backptr, Bool indirect )
{
    if (!backptr)
        return 1;
//    dyalog_printf( "void? [%d]\n", B_TYPE( backptr )); /* trace */
    switch (B_TYPE( backptr )) {
        case f_or:
            return forest_void( B_CAR( backptr), indirect)
                && forest_void( B_CDR( backptr), indirect);
            break;
        case f_join:
        case f_rand:
            return 0;
            break;
        case f_and:
//            dyalog_printf("FAND %x\n",B_CDR(backptr)); /* trace */
            if (!indirect)
                return 0;
            if (!forest_void(((tabobj_t)B_CAR(backptr))->backptr,indirect))
                return 0;
            return (!B_CDR(backptr) || forest_void(((tabobj_t)B_CDR(backptr))->backptr,0));
            break;
        case f_trace:
            return forest_void( B_CDR(backptr),indirect);
            break;
        case f_indirect:
        {
            long info=0;
            obj_t entry=0;
            find_or_add_indirect(backptr,&info,0,&entry);
            return info & INDIRECT_VOID;
            break;
        }
        default:
            return 1;
            break;
    }
}

Bool
forest_indirect( obj_t backptr )
{
//    dyalog_printf( "ind? [%d]\n", B_TYPE( backptr )); /* trace */
    
    switch (B_TYPE( backptr )) {
        case f_or:
            return forest_indirect( B_CAR( backptr ) ) && forest_indirect( B_CDR( backptr));
            break;
        case f_indirect:
            return 1;
            break;
        case f_trace:
            return forest_indirect( B_CDR( backptr ));
            break;
        default:
            return 0;
            break;
    }
}

static Bool
forest_display_body( obj_t backptr, Bool balance )
{
    Bool forest=0;
    fol_t trace = (fol_t) 0;
    obj_t b;

    backptr = forest_remove_duplicates(backptr);
        
//    dyalog_printf( "body [%d]\n", B_TYPE(backptr) ); /* trace */
    switch (B_TYPE( backptr )) {
        case f_call:
        case f_init:
            
            return 0;
            break;

        case f_or:
            if (!balance) dyalog_printf( "( " );
            if (!forest_display_body( B_CAR( backptr ), 1)) dyalog_printf( " e " );
            dyalog_printf( " | " );
            if (!forest_display_body( B_CDR( backptr ), 1)) dyalog_printf( " e " );
            if (!balance) dyalog_printf( " )" );
            return 1;
            break;
    
        case f_join:
            if (! forest_void( ((tabobj_t)B_CAR(backptr))->backptr,0 )){
                dyalog_printf( "%d", get_index( (tabobj_t) B_CAR(backptr) ));
                forest=1;
            } 
            if (! forest_void( ((tabobj_t)B_CDR(backptr))->backptr,0 )){
                if (forest) {dyalog_printf(" ");}
                dyalog_printf( "%d", get_index( (tabobj_t) B_CDR(backptr) ));
                forest |= 1;
            } 
            return forest;
            break;

        case f_and:
//            dyalog_printf("CAT %d\n",(int)B_CAR(backptr)); /* trace */
            b=((tabobj_t)B_CAR(backptr))->backptr;
            if (B_TYPE(b) == f_trace){
                trace = (fol_t) B_CAR(b);
                b=B_CDR(b);
            }
            if (forest_display_body(b,0)){
                forest = 1;
                dyalog_printf( " " );
            }
//            dyalog_printf("CDR %d\n",(int)B_CDR(backptr)); /* trace */
            if ((tabobj_t)B_CDR(backptr)){
                if (forest_indirect(((tabobj_t)B_CDR(backptr))->backptr)) {
                    tabobj_t o = (tabobj_t)B_CDR(backptr);
                    if (trace) {
                        dyalog_printf( "[%&f]", trace);
                        forest=1;
                    }
                    forest=forest_display_body( o->backptr,0);
                } else if (! forest_void( ((tabobj_t)B_CDR(backptr))->backptr,0 )){
                    if (trace) {
                        dyalog_printf( "[%&f]", trace);
                        forest=1;
                    }
                    dyalog_printf( "%d", get_index( (tabobj_t) B_CDR(backptr) ));
                    forest=1;
                } else if (trace) {
                        // This case should be an error
//                    dyalog_printf("HERE ");
                }                
            }
            return forest;
            break;

        case f_rand:
            if (! forest_void( ((tabobj_t)B_CAR(backptr))->backptr,0 )){
                dyalog_printf( "%d", get_index( (tabobj_t) B_CAR(backptr) ));
                forest=1;
            } else if (forest_indirect(((tabobj_t)B_CAR(backptr))->backptr)) {
                if (forest_display_body( ((tabobj_t)B_CAR(backptr))->backptr, 0))
                    forest=1;
            }
            if (forest)
                dyalog_printf( " " );

            if (forest_display_body( ((tabobj_t)B_CDR(backptr))->backptr, 0)){
                forest |= 1;
            }
            return forest;
            break;

        case f_indirect: 
        {
            long info=0;
            obj_t entry=0;
            long i=find_or_add_indirect((obj_t)backptr,&info,INDIRECT_FOUND,&entry);
//            dyalog_printf("{%d:%d}",i,info); /* trace */
            if (!(info & INDIRECT_VOID)){
                if (info & INDIRECT_CYCLE) {
                        /* non empty cycling indirect => need to use #x
                         * notation to print the forest */
                    dyalog_printf("#%d",i);
                    if (!(info & INDIRECT_FOUND)){
                        dyalog_printf( "{ " );
                        CDR(CAR(entry)) = (obj_t) (((long) CDR(CAR(entry))) | INDIRECT_FOUND);
                        forest=forest_display_body( B_CAR( backptr ), 1);
                        dyalog_printf( " }" );
                    } else {
                        dyalog_printf(" ");
                    }
                } else {
                        /* non empty non cycling indirect => can expand
                         * content and no need for #x (except for compactness) */
                    forest=forest_display_body( B_CAR( backptr ), 0);
                }
            }
        }
        return forest; 
        break;

        default:
            return 0;
            break;
    }
}

void
forest_display( obj_t solutions )
{
  dyalog_printf( "\n"
	  "----------------------------------------------------------------------\n"
	  "Shared Forest\n"
	  );

  seen = solutions; 
  indirect = BNIL;
  
  for( ; PAIRP( solutions ) ; solutions = CDR( solutions ) ){
    tabobj_t o = (tabobj_t) CAR(solutions);
    struct sfol s = { t:o->seed->model, k:Trail_Object( o )};
//    fkey_t k = Trail_Object( o );
    long  i = get_index(o);
//           dyalog_printf( "\n%&c\n\t%d <-- ", o->seed->model, k, i ); /* trace */
    DyALog_Assign_Display_Info( &s );
    dyalog_printf( "\n%&c\n\t%d <-- ", s.t, s.k, i );
    forest_display_body( o->backptr, 0);
    Untrail_Object;
  }

  dyalog_printf( "\n" );	  

}


/**********************************************************************
 * Counting derivation trees
 **********************************************************************/

typedef struct count {
  tabobj_t key;
  double    count;
  struct count *next;
} *count_t;

static count_t solutions_count;
static long node_counter=0;

static double forest_count_head( tabobj_t );

static double
forest_count_body( obj_t backptr )
{
  
    if (forest_void( backptr,0 ))
        return 1;
    else
        switch (B_TYPE( backptr )) {
            case f_or:
                return forest_count_body( B_CAR( backptr )) + forest_count_body( B_CDR( backptr ));
                break;
            case f_join:
                return forest_count_head( (tabobj_t) B_CAR( backptr )) *
                    forest_count_head( (tabobj_t) B_CDR( backptr ));
                break;
            case f_and:
                return forest_count_body( ((tabobj_t) B_CAR( backptr ))->backptr) *
                    forest_count_head( (tabobj_t) B_CDR( backptr ));
                break;
            default:
                return 1.0 ;
                break;
        }
}

static double
forest_count_head( tabobj_t o )
{
  count_t *c = &solutions_count;
  
  for(; *c && (*c)->key != o; c = &(*c)->next );
  
  if (!*c) {
    count_t entry = (count_t) GC_MALLOC( sizeof( struct count ) );
    *c = entry;
    entry->key  = o;
    entry->count = 0.0;
    entry->count = forest_count_body( o->backptr);
    ++node_counter;
  }
  
  return (*c)->count;
}

void
forest_count( obj_t solutions )
{
  double derivation_counter=0.0;
  dyalog_printf( "\n"
		 "----------------------------------------------------------------------\n"
		 "\n" );

  seen = solutions; 

  for( ; PAIRP( solutions ) ; solutions = CDR( solutions ) ){
    tabobj_t o = (tabobj_t) CAR(solutions);
    fkey_t k = Trail_Object( o );
    double nb = forest_count_head(o);
    derivation_counter = derivation_counter + nb;
    dyalog_printf( "%g derivations for %&c\n",nb, o->seed->model, k);
    Untrail_Object;
  }

  dyalog_printf( "Used %d nodes for %g derivations\n\n", node_counter, derivation_counter);
}





