/*************************************************************
 * $Id: rt.c 471 2005-02-23 17:33:30Z clerger $
 * Copyright (C) 1997, 2002, 2003, 2004, 2005 Eric de la Clergerie
 * ------------------------------------------------------------
 *
 * Run Time   --
 *
 * ------------------------------------------------------------
 * Description
 *      DyALog runtime.
 *
 * ------------------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "libdyalog.h"
#include "builtins.h"

extern obj_t dyalog_solutions;
extern fol_t _dyalog_argv;
extern int _dyalog_initialization_stamp;

extern void internal_init();
extern void folsmb_exec_declaration( fol_t, fkey_t );
extern tabobj_t rt_install_database( fol_t, fkey_t );
extern void File_Include( const char *, const obj_t);
extern void parse_from_stm(int, obj_t);
extern void dyalog_loop();
extern void forest_display( obj_t );
extern void forest_count( obj_t );
extern Bool Close_1(Sproto);
extern void initialization_table();

extern int server_accept( int );
extern int server_create();

typedef enum {Forest_Nothing=0, Forest_Display, Forest_Count} Forest_Action;

static char string_getc(FILE *);
static void usage( char *);
static void file_db( char *);
static void stringlex( FILE*, char port_getc(FILE*), char );

extern void add_path(const char *);

static char *command;

/**********************************************************************
 * file_db
 **********************************************************************/

void
db_clause( obj_t handler, obj_t o, obj_t kk )
{
  fol_t t = (fol_t) o;
  fkey_t Sk(t) = (fkey_t) kk;
  
  if (FOLCMPP(t) && FOLCMP_FUNCTOR(t) == find_folsmb( ":-", 1 ))
    folsmb_exec_declaration( FOLCMP_REF(t,1), Sk(t) );
  else
    rt_install_database( t, Sk(t) );
}

DEFINE_STATIC_PROCEDURE(db_clause_handler,db_clause_closure,db_clause,0L,2);

void
file_db( char *filename )
{
  File_Include( filename, db_clause_handler );
}

/**********************************************************************
 * Reading from a string
 **********************************************************************/

char
string_getc( FILE *port )
{
  char **handle= (char **) port;
  char c = **handle;
  
  if (c) (*handle)++;

  return c;
}

static int
decode_verbose_level( const char * level )
{

    if (strcmp(level,"dyam") == 0)
        return V_DYAM;

    if (strcmp(level,"index") == 0)
        return V_INDEX;

    if (strcmp(level,"share") == 0)
        return V_SHARE;

    if (strcmp(level,"term") == 0)
        return V_TERM;

    if (strcmp(level,"tfs") == 0)
        return V_TFS;

    if (strcmp(level,"builtins") == 0)
        return V_BUILTINS;

    if (strcmp(level,"low") == 0)
        return V_LOW;
    
    return V_DYAM + V_INDEX + V_SHARE + V_TFS + V_BUILTINS + V_LOW;
}

void set_verbose_level(int level)
{
    verbose_level = level;
}

int get_verbose_level()
{
    return verbose_level;
}

/**********************************************************************
 * Run
 **********************************************************************/


static void bye() 
{
    FREE_ALLOCATION();
}

int
main( int argc, char *argv[])
{
  int i=1;
  Forest_Action forest = Forest_Nothing ;
  int server = 0;

#ifdef HAVE_ATEXIT  
  atexit(bye);                 /* registering exit function */
#endif
  
  verbose_level=0;
  
//  setlocale(LC_CTYPE, "");        /* Set locale for lexer */
     
  command = argv[0];

  /* Internal Initialization : bigloo + DyALog */
  internal_init();

  /* Parsing command line up to -a or end of line */
  for(; i < argc ;) {
    char *arg = argv[i++];

    if (strcmp(arg,"-h") == 0 || strcmp(arg,"-help") == 0)
      usage( NULL );

    if (strcmp(arg,"-v") == 0  || strcmp(arg,"-verbose") == 0) {
        if (i == argc) {
            usage( arg );
        } else {
            verbose_level |= decode_verbose_level(argv[i++]);
            continue;
        }
    }
    
    if (strcmp(arg,"-db") == 0){
      if (i == argc)
	usage( arg );
      else {
	file_db( argv[i++] );
	continue;
      }}

    if (strcmp(arg,"-forest") == 0) {
      forest = Forest_Display;
      continue;
    }

    if (strcmp(arg,"-fcount") == 0) {
      forest = Forest_Count;
      continue;
    }
      
    if (strcmp(arg,"-slex") == 0){
      if (i == argc) usage( arg );
      else {
	stringlex( (FILE*) &argv[i++], string_getc, '\0');
	continue;
      }}

    if (strcmp(arg,"-flex") == 0) {
      if (i == argc) usage( arg );
      else {
	char *filename = argv[i++];
	if (strcmp(filename,"-") == 0) {
	  stringlex( stdin, (char (*)(FILE*)) fgetc, EOF);
	} else {
	  FILE *file = fopen( filename, "r" );
	  stringlex( file, (char (*)(FILE*)) fgetc, EOF );
	  fclose(file);
	}
	continue;
      }
    }

    if (strcmp(arg,"-I") == 0) {
        if (i == argc)
            usage( arg );
        else {
            add_path( (char *)argv[i++] );
            continue;
        }
    }

    
    if (strcmp(arg,"-server") == 0) {
      /* server mode: to be refined! */
      server = server_create();
      continue;
    }

    if (strcmp(arg,"-a") == 0 || strcmp(arg,"--") == 0)
        break;                   /* remaining options specific to DyALog application */

    if ( arg[0] == '-' && arg[1] != '\0' ){
//      usage( arg );
        i--;
        break;                  /* remaining options specific to DyALog application */
    }
    
				/*  default */

    file_db( arg );

  }

				/* Build options for DyALog */

  for(_dyalog_argv=FOLNIL ;
      i < argc ;
      _dyalog_argv = folcmp_create_pair( Create_Atom(argv[--argc]), _dyalog_argv )
      );

  do {

    /* In server mode : read next database */
    if (server) {
      stm_output=server_accept( server );
      parse_from_stm(stm_output,db_clause_handler);
    }

    /* Main Initialization: defined in the generated MA file */
    main_initialization();

    /* Loop */
    dyalog_loop();
    
    /* Post Processing */
    switch (forest) {
    case Forest_Display:
      forest_display( dyalog_solutions );
      break;
    case Forest_Count:
      forest_count( dyalog_solutions );
      break;
    default:
      break;
    }
   
    if (server) {		/* Do some cleaning before next round */
      Close_1( DFOLINT(stm_output), 0 );
      _dyalog_initialization_stamp = (_dyalog_initialization_stamp + 1) % 2;
      dyalog_solutions = BNIL;
      initialization_table();
    }
    
  } while (server);

  if (server)
    close(server);
                                        /* Exit */
  
   return EXIT_SUCCESS;
}

/**********************************************************************
 * usage
 **********************************************************************/

void
usage( char *error )
{
  if (error) {
    printf( "Error with option %s\n"
	    "------------------------------------------------------------\n",
	    error );
  }
  printf(  "Usage for command %s\n"
	   "\t-h -help            -- this help\n"
           "\t-v <level>          -- trace for <level> in dyam, index, share, or all\n"
           "\t-verbose <level>    -- same as -v\n"
	   "\t-db <filename>      -- load database filename\n"
           "\t-I <path>           -- add <path> to DyALog search path\n"
	   "\t-server             -- enter server mode\n"
	   "\t-forest             -- display the shared forest\n"
	   "\t-fcount             -- count number of derivations per answer\n"
	   "\t-slex <string>      -- parse from <string>\n"
	   "\t-flex <filename>    -- parse from <filename>\n"
	   "\t-a <args>           -- all remaining arguments given to DyALog\n"
           "\t-- <args>           -- all remaining arguments given to DyALog\n"
	   , command);
  exit(error ? EXIT_FAILURE : EXIT_SUCCESS);
}

/**********************************************************************
 * stringlex
 *     read a char file (or string) and install all chars as connect facts
 **********************************************************************/

void
stringlex( FILE *port, char port_getc(FILE *), char port_eof )
{
  int i=0;
  char c;
  fol_t f= find_folsmb("C",3);
  fkey_t k = Key0;
  fol_t t;
  for(; (c=port_getc(port)) != port_eof; )
    if (c != '\n') {
      FOLCMP_WRITE_START( f , 3);
      FOLCMP_WRITE( DFOLINT( i ) );
      FOLCMP_WRITE( DFOLCHAR( c ) );
      FOLCMP_WRITE( DFOLINT( i+1 ));
      t = FOLCMP_WRITE_STOP;
      rt_install_database( t, k );
      i++;
    }
  rt_install_database( folcmp_create_unary( "N", DFOLINT( i )), k);
}
    
