/************************************************************
 * $Id: tree-indexing.c 566 2006-11-06 17:20:50Z clerger $
 * Copyright (C) 1997, 2003, 2004, 2006 Eric de la Clergerie
 * ------------------------------------------------------------
 *
 *   TreeIndexing --
 *
 * ------------------------------------------------------------
 * Description
 *      Term Indexing using Decision Trees
 *
 * Note
 *      the **DyALog Builtin** comments (for instance Unif_Retrieve2)
 *      are used by a perl extractor to build builtin.ma (see Makefile)
 * ------------------------------------------------------------
 */

#include <string.h>

//#define VERBOSE

#include "libdyalog.h"
#include "builtins.h"
#include "hash.h"
#include "tree-indexing.h"

/* Following definition moved into rt.h */
/* typedef struct obj_cell *cell_t; */

struct obj_cell {
  tabobj_t object;
  cell_t   next;		
};

#define OBJ_CELL_SIZE             sizeof( struct obj_cell )

typedef enum { TABLE=1, LIST } entry_kind;

typedef struct {
    fol_t   key;
    entry_kind kind;
    int     depth;
    int     son_depth;
    cell_t list;
} table_entry_t, *TableEntry;

static table_entry_t dyalog_entry =  { DFOLINT(0), TABLE, 0, -1, (void *) 0};
static TableEntry dyalog_table = &dyalog_entry;

#define START_TABULATION_SIZE      4
#define TYPE_SIZE                  sizeof( table_entry_t )

#define VarKey FOLVAR_FROM_INDEX(0)

#define HOLE_P( A )  ( A == FOLHOLE || FOL_DEREFP(A)  )

/**********************************************************************
 * Initialization
 **********************************************************************/

void initialization_table()
{
  dyalog_entry.list = (void *) Hash_Alloc_Table( START_TABULATION_SIZE, TYPE_SIZE);
}

/**********************************************************************
 * Insertion
 **********************************************************************/

static
table_entry_t *
add( SP(A), int depth, const Bool brother, char *table)
{
  fol_t  key;
  Bool nonterminal = brother ;
  table_entry_t tmp, *entry;

  Deref(A);

  if (HOLE_P(A)) {
    key = VarKey;
  } else if (FOLCMPP(A)) {
    nonterminal |= (depth < INDEXATION_MAX_DEPTH );
    key = FOLCMP_FUNCTOR(A);
  } else
    key = A;

  V_LEVEL_DISPLAY( V_INDEX, "\tctr=%d, brother=%d, key=%&f\n",depth,brother,key);
  
  tmp = (table_entry_t) { key: key ,
				 kind: ( nonterminal ? TABLE : LIST ),
				 depth: depth,
				 son_depth: depth,
				 list: (void *) 0 };
  entry = (TableEntry)  Hash_Insert( table, (char *) &tmp, (Bool) 0);

  if ( (!entry->list) && nonterminal ) {
    /* entry is new, void and should hold a table */
    entry->list = (void *) Hash_Alloc_Table( START_TABULATION_SIZE, TYPE_SIZE);
  }

  if ( depth++ < INDEXATION_MAX_DEPTH && FOLCMPP(A) && (!FOLCMP_DEREFP(A))) {
    /* Index on arguments of A  */
    fol_t *arg  = &FOLCMP_REF(A,1);
    fol_t *stop = arg +  FOLCMP_ARITY(A) -1;

    entry->son_depth = depth;	            /* depth of entry sons is old depth+1  */

    for (;
	 arg <= stop ;
	 entry = add( *arg, k_A, depth, (arg != stop) || brother , (char *) entry->list), arg++ );

    entry->son_depth = depth-1;             /* depth of last entry's sons is old depth */
  }

  return entry;

}

void
tabulation_add( const tabobj_t o )
{
    char *table = (char *)dyalog_table->list;
    fkey_t k_A  = Trail_Object( o );
        //  fol_t A     = TABOBJ_SUBS( o );
    fol_t A     = TABOBJ_MODEL( o );
    
    V_LEVEL_DISPLAY( V_INDEX, "inserting %&s\n", A, k_A );

    {
        TableEntry anchor = (TableEntry)  add(A,k_A,0,0,table);
        cell_t cell = (cell_t) GC_MALLOC( OBJ_CELL_SIZE );
        V_LEVEL_DISPLAY( V_INDEX, "I am here\n" );
        cell->object = o;
        cell->next   = anchor->list;
        anchor->list = cell;
        o->delete_info = (void *) &(anchor->list);
        if ((cell->next)){
            cell->next->object->delete_info=(void *)&(cell->next);
        }
    }
    
    V_LEVEL_DISPLAY( V_INDEX, "Done insertion\n" );
    Untrail_Object;
}
    
void                                 /* Dyam Instruction */
Dyam_Tabule()                        /* object is built and stored in R_OBJECT */
{
    tabulation_add(R_OBJECT);
}

/**********************************************************************
 * Retrieval Closure mechanism
 **********************************************************************/

extern void C_To_DyALog();
extern void DyALog_Fail();
extern void DyALog_To_C();
extern void Closure_Apply();

Bool
closure_apply(const cell_t list)
// **DyALog Builtin** Closure_Apply :- closure_apply(0)
{
    tabobj_t object = list->object;
    ClosedFn fn = * (ClosedFn *) IP->closure;
    if (list->next) {
        LVALUE_X(0)= REG_VALUE(list->next);
        update_choice(Closure_Apply,1);
    } else {
        Dyam_Remove_Choice();
    }    
    V_LEVEL_DISPLAY( V_INDEX, "closure apply\n" );
    V_LEVEL_DISPLAY( V_INDEX, "\tApplying closure to %&f\n",TABOBJ_MODEL(object));
    return (*fn)( object, IP->closure );
}

Bool
closure_apply_start(const cell_t list) 
{
    tabobj_t object = list->object;
    ClosedFn fn = * (ClosedFn *) IP->closure;
    if (list->next) {
        LVALUE_X(0)=REG_VALUE(list->next);
        Dyam_Full_Choice(Closure_Apply,1);
    } else {
    }    
    V_LEVEL_DISPLAY( V_INDEX, "closure apply\n" );
    V_LEVEL_DISPLAY( V_INDEX, "\tApplying closure to %&f\n",TABOBJ_MODEL(object));
    return (*fn)( object, IP->closure );
}

/**********************************************************************
 *  Retrieval Resume and Undo mechanism
 **********************************************************************/

static HashScan *
Start_HashScan()
{
    PUSH_TRAIL_BOX(Trailed_HashScan,box);
    box->type = HASH_SCAN;
    return &(box->scan);
}

static void
Delete_Topmost_Trailed_Bloc()
{
    POP_TRAIL_BOX(LVALUE_C_TRAIL_TOP);
}

static void
Start_Retrieval(SP(A), const Closure closure)
{
    PUSH_TRAIL_BOX(Resume,box);
    box->type = INDEXATION;
    box->prev = IP;
    box->closure = closure;
    box->depth = 0;
    box->A = A;
    box->info[0] = (struct ResumeInfo) { arg:&(box->A), stop:&(box->A)+1, k: Sk(A) };
    LVALUE_IP = REG_VALUE(box);
}

static void
Retrieval_To_DyALog(const fun_t fun, SP(A), const Closure closure )
{
    Dyam_Full_Choice(DyALog_To_C,1);
    LVALUE_X(0)=REG_VALUE(dyalog_table);
    Start_Retrieval(S(A),closure);
    LVALUE_R_CP= (fun_t) Adjust_Address(DyALog_Fail);
//    R_BC=R_B;
    LVALUE_R_P= (fun_t) Adjust_Address(fun);
    C_To_DyALog();
}

static void
Retrieval_To_DyALog_With_Cut(const fun_t fun, SP(A), const Closure closure )
{
    Dyam_Full_Choice(DyALog_To_C,1);
    LVALUE_X(0)=REG_VALUE(dyalog_table);
    Start_Retrieval(S(A),closure);
    LVALUE_R_CP=(fun_t) Adjust_Address(DyALog_Fail);
    LVALUE_R_BC=R_B;
    LVALUE_R_P= (fun_t) Adjust_Address(fun);
    V_LEVEL_DISPLAY(V_DYAM,"Start Retrieval_To_DyALog\n");
    C_To_DyALog();
}

static void
resume_forward()	/* Advance in resume and possibly pop */
{
    int pop=IP->depth;
    
    V_LEVEL_DISPLAY( V_INDEX, "\tforward ..." );

    while( ++(IP->info[IP->depth].arg) == (IP->info[IP->depth].stop )
           && (IP->depth > 0) )
        IP->depth--;

    pop -= IP->depth;

    V_LEVEL_DISPLAY( V_INDEX, " -%d, depth is now %d\n", pop, IP->depth );

    if ((void *)IP < (void *) (R_B->trail)) {
        PUSH_TRAIL_BOX(Undo,undo);        
        undo->type = FORWARD;
        undo->value.forward = pop;
    }
    
}

static void
resume_push( SP(A) )   /* Push A in resume where A is a compound term */
{
    if ((IP->depth < INDEXATION_MAX_DEPTH) && FOLCMPP(A)) {
        fol_t *start=&FOLCMP_REF(A,1);
        fol_t *stop= start + FOLCMP_ARITY(A);
        int depth = ++IP->depth;
        
        V_LEVEL_DISPLAY( V_INDEX, "\tpushing %&s to depth %d\n", A, k_A, depth );
        if ((void *) IP < (void *) (R_B->trail)) {
            PUSH_TRAIL_BOX(Undo,undo); 
            undo->type = UPWARD;
            undo->value.upward =  (struct Upward)
                { stop: IP->info[depth].stop, k: IP->info[depth].k };
        }
        IP->info[depth] = (struct ResumeInfo) { arg:start, stop:stop, k:k_A };
        
    } else 
        resume_forward();
}

/**********************************************************************
 * Unification-based Retrieval
 **********************************************************************/

static Bool blind_retrieve3( const TableEntry );
static Bool blind_retrieve( const char * );

Bool unif_retrieve2( TableEntry );
Bool unif_retrieve3( TableEntry, Sproto);

extern void Blind_Retrieve2();
extern void Unif_Retrieve3();
extern void Unif_Retrieve2();

Bool
blind_retrieve2(const TableEntry entry, HashScan *scan)
// **DyALog Builtin** Blind_Retrieve2 :- blind_retrieve2(0,1)
{
    TableEntry next = (TableEntry)Hash_Next(scan);
    V_LEVEL_DISPLAY( V_INDEX,"in blind_retrieve2 entry key %&f\n",entry->key);
    if (next) {
        LVALUE_X(0)= REG_VALUE(next);
        LVALUE_X(1)= REG_VALUE(scan);
        update_choice(Blind_Retrieve2,2);
        V_LEVEL_DISPLAY( V_INDEX,"Setting choice point in blind_retrieve2 %&f\n%&t\n",next->key);
    } else {
        Dyam_Remove_Choice();
        Delete_Topmost_Trailed_Bloc();
    }        
    return blind_retrieve3(entry);
}

Bool
blind_retrieve2_start(const TableEntry entry, HashScan *scan)
{
    TableEntry next = (TableEntry)Hash_Next(scan);
    V_LEVEL_DISPLAY( V_INDEX,"starting blind_retrieve2 entry key %&f\n",entry->key);
    if (next) {
        LVALUE_X(0)= REG_VALUE(next);
        LVALUE_X(1)= REG_VALUE(scan);
        Dyam_Full_Choice(Blind_Retrieve2,2);
        V_LEVEL_DISPLAY( V_INDEX,"Setting choice point in blind_retrieve2 %&f\n%&t\n",next->key);
    } else {
        Delete_Topmost_Trailed_Bloc();
    }        
    return blind_retrieve3(entry);
}

static Bool
blind_retrieve3(const TableEntry entry)
{
    if (entry->kind == LIST){
        V_LEVEL_DISPLAY( V_INDEX, "blind_retrieve3 LIST\n");
        return entry->list && closure_apply_start(entry->list);
    } else if (entry->son_depth > IP->depth) {
        V_LEVEL_DISPLAY( V_INDEX, "blind_retrieve3 TOO DEEP\n");
        return blind_retrieve((char *) entry->list );
    } else {
        V_LEVEL_DISPLAY( V_INDEX, "Restarting from blind_retrieve at depth %d\n", IP->depth);
        return unif_retrieve2(entry);
    }
}

static Bool
blind_retrieve( const char *table )
{
    HashScan *scan=Start_HashScan();
    TableEntry entry = (TableEntry)Hash_First(table,scan);

    return entry && blind_retrieve2_start(entry,scan);
}

Bool
unif_retrieve2( TableEntry entry)
// **DyALog Builtin** Unif_Retrieve2 :- unif_retrieve2(0)
{
    while (1) {
        V_LEVEL_DISPLAY( V_INDEX, "Unif retrieve\n" );
            // assert( !entry || entry->depth == depth );
        if (entry->kind == LIST) {
            V_LEVEL_DISPLAY( V_INDEX, "\tentry info key=%&f, depth=%d, kind=%d\n", entry->key ,
                       entry->depth, entry->kind );
            return entry->list && closure_apply_start(entry->list);
        } else if ( IP->depth >= 0) {
            int depth = IP->depth;
            fol_t A = *(IP->info[depth].arg);
            fkey_t k_A = IP->info[depth].k;
            char *table = (char *) entry->list;
            fol_t  key ;
            TableEntry e1,e2;

            V_LEVEL_DISPLAY( V_INDEX,"toto\n");
            
            Deref(A);
      
            key  = FOLCMPP(A) ? FOLCMP_FUNCTOR(A) : A;

            V_LEVEL_DISPLAY( V_INDEX, "\tentry info key=%&f, depth=%d, kind=%d\n", entry->key ,
                       entry->depth, entry->kind );
            V_LEVEL_DISPLAY( V_INDEX, "\tcurrent term is %&f\n", A );
            
            if (HOLE_P(A)) {                   /* try recursively all entries of table */
                V_LEVEL_DISPLAY( V_INDEX, "\tblind_retrieve\n");
                resume_forward();
                return blind_retrieve(table);
            }
            

                           /* Try variable entry in table */
            e1 = (TableEntry) Hash_Find(table,(long)VarKey);
            e2 = (TableEntry) Hash_Find(table,(long) key);

            if (e1) {
                if (e2) {
                    V_LEVEL_DISPLAY( V_INDEX,"Setting choice point in unif_retrieve2 %&f\n",e2->key);
                    LVALUE_X(0) = REG_VALUE(e2);
                    LVALUE_X(1)=  REG_VALUE(A);
                    LVALUE_X(2)=  REG_VALUE(Sk(A));
                    Dyam_Full_Choice(Unif_Retrieve3,3);
                }
                V_LEVEL_DISPLAY( V_INDEX,"<>E1\n");
                entry= e1;
                resume_forward(); 
            } else if (e2) {
                V_LEVEL_DISPLAY( V_INDEX,"<>E2\n");
                entry= e2;
                resume_push(S(A));
            } else {
                V_LEVEL_DISPLAY( V_INDEX,"Fail\n");
                Fail;
            }
        }
    }
}

Bool
unif_retrieve3(TableEntry entry, SP(A))
// **DyALog Builtin** Unif_Retrieve3 :- unif_retrieve3(0,1,2)
{
    Dyam_Remove_Choice();
    resume_push(S(A));
    return unif_retrieve2(entry);
}

void
unif_retrieve( SP(A), const Closure closure )
{
    V_LEVEL_DISPLAY( V_INDEX,"UNIF RETRIEVE %&s\n",S(A));
    if (dyalog_table) {
        Retrieval_To_DyALog((fun_t)Unif_Retrieve2,S(A),closure);
        V_LEVEL_DISPLAY( V_INDEX,"END UNIF RETRIEVE\n");
    }
}

Bool
unif_retrieve_alt( SP(A), const Closure closure )
{
    if (dyalog_table) {
        Start_Retrieval(S(A),closure);
        return unif_retrieve2(dyalog_table);
    } else {
        Fail;
    }
}


/**********************************************************************
 * Subsumption-based  Retrieval
 **********************************************************************/

Bool subs_retrieve2(TableEntry entry);
Bool subs_retrieve3(TableEntry entry, Sproto);

extern void Subs_Retrieve3();
extern void Subs_Retrieve2();

Bool
subs_retrieve2(TableEntry entry)
// **DyALog Builtin** Subs_Retrieve2 :- subs_retrieve2(0)
{
    while (1) {
        V_LEVEL_DISPLAY( V_INDEX, "Subs retrieve\n" );
        
            // assert( !entry || entry->depth == depth );
        
        if (entry->kind == LIST) {
            V_LEVEL_DISPLAY( V_INDEX, "\tentry info key=%&f, depth=%d, kind=%d\n", entry->key ,
                       entry->depth, entry->kind );
            return entry->list && closure_apply_start(entry->list);
        } else if ( IP->depth >= 0 ) {
            int depth=IP->depth;
            fol_t A = *(IP->info[depth].arg);
            fkey_t k_A = IP->info[depth].k;
            char *table = (char *) entry->list;
            fol_t key;
            TableEntry e1;
            TableEntry e2=0;
        
            Deref(A);

            key = FOLCMPP(A) ? FOLCMP_FUNCTOR(A) : A;
        
            V_LEVEL_DISPLAY( V_INDEX, "\tentry info key=%&f, depth=%d, kind=%d\n", entry->key , entry->depth, entry->kind );
            V_LEVEL_DISPLAY( V_INDEX, "\tcurrent term is %&f\n", A );

            e1 = (TableEntry) Hash_Find(table,(long)VarKey);
            if (!HOLE_P(A)) e2 = (TableEntry) Hash_Find(table,(long) key);
                
                /* Try variable entry in table */
            if (e1) {
                if (e2) {
                    V_LEVEL_DISPLAY( V_INDEX,"Setting choice point in unif_retrieve2 %&f\n",e2->key);
                    LVALUE_X(0) = REG_VALUE(e2);
                    LVALUE_X(1) = REG_VALUE(A);
                    LVALUE_X(2) = REG_VALUE(Sk(A));
                    Dyam_Full_Choice(Subs_Retrieve3,3);
                }
                entry=e1;
                resume_forward();
            } else if (e2) {
                entry=e2;
                resume_push(S(A));
            } else {
                Fail;
            }
        }
    }
}

Bool
subs_retrieve3(TableEntry entry, SP(A))
// **DyALog Builtin** Subs_Retrieve3 :- subs_retrieve3(0,1,2)
{
    Dyam_Remove_Choice();
    resume_push(S(A));
    return subs_retrieve2(entry);
}

void
subs_retrieve( SP(A), const Closure closure )
{
    V_LEVEL_DISPLAY( V_INDEX,"SUBS RETRIEVE %&s\n",S(A));
    if (dyalog_table) {
        Retrieval_To_DyALog((fun_t)Subs_Retrieve2,S(A),closure);
    }
}

/**********************************************************************
 * Variance-Based Retrieval
 ***********************************************************************/

extern void Variance_Retrieve2();

Bool
variance_retrieve2( TableEntry entry )
        // **DyALog Builtin** Variance_Retrieve2 :- variance_retrieve2(0)
{
    while (entry) {
        
        V_LEVEL_DISPLAY( V_INDEX, "Variance retrieve\n" );
            // assert( !entry || entry->depth == depth );

        if (entry->kind == LIST) {
            V_LEVEL_DISPLAY( V_INDEX, "\tentry info key=%&f, depth=%d, kind=%d\n", entry->key ,
                       entry->depth, entry->kind );
            return entry->list && closure_apply_start(entry->list);
        } else if ( IP->depth >= 0 ) {
            int depth=IP->depth;
            fol_t A = *(IP->info[depth].arg);
            fkey_t k_A = IP->info[depth].k;
            char *table = (char *) entry->list;
            fol_t key;
 
            Deref(A);

            key = FOLCMPP(A) ? FOLCMP_FUNCTOR(A) : A;
 
            V_LEVEL_DISPLAY( V_INDEX, "\tentry info key=%&f, depth=%d, kind=%d\n", entry->key , entry->depth, entry->kind );
            V_LEVEL_DISPLAY( V_INDEX, "\tcurrent term is %&f\n", A );

                /* Try variable entry in table */

            if (HOLE_P(A)) {
                entry = (TableEntry) Hash_Find(table,(long)VarKey);
                resume_forward();
            } else {
                entry = (TableEntry) Hash_Find(table,(long) key);
                resume_push(S(A));
            }
        }
    }
    Fail;
}


void
variance_retrieve( SP(A), const Closure closure )
{
    V_LEVEL_DISPLAY( V_INDEX,"VARIANCE RETRIEVE %&s\n",S(A));
    if (dyalog_table)
        Retrieval_To_DyALog((fun_t)Variance_Retrieve2,S(A),closure);
}

/**********************************************************************
 * Standard DyALog Indexation-based functions
 **********************************************************************/

				/* Application  */

typedef struct application_closure {
    void     *fn;
    long      id;
    fol_t     type;
    fun_t     code;
    tabobj_t  transition;
    fkey_t    k;
} application_closure;

Bool treat_item( tabobj_t, long, fol_t, fun_t, tabobj_t, fkey_t );

static Bool
closed_application( tabobj_t obj, struct application_closure *argv )
{
    return treat_item(obj, argv->id, argv->type, argv->code, argv->transition,argv->k);
}


Bool                            /* DYAM INSTRUCTION Dyam_Apply */
Dyam_Apply(int item_comp_id, int trans_comp_id)
{
    tabseed_t seed = R_TRANS->seed;
    void **comptab = ((void **)&seed->application_code+1);
    trans_comp_t comp = ((trans_comp_t) comptab)+trans_comp_id;
    fol_t model = comp->model;
    fun_t code = comp->code;
    fol_t type = seed->type;
    Closure closure = (Closure) PUSH_TRAIL_BLOCK(sizeof(struct application_closure));
    *(struct application_closure *)closure = (struct application_closure)
        { (void *)&closed_application,
              item_comp_id,
              type,
              code,
              R_TRANS,
              R_TRANS_KEY };
    
    V_LEVEL_DISPLAY( V_DYAM,"  [%&e] Apply %&s %d\n",model,R_TRANS_KEY,item_comp_id);
    return unif_retrieve_alt(model,R_TRANS_KEY,closure);
}


				/* Completion */

struct completion_closure {
    void *fn;
    long  id;
    fol_t type;
    tabobj_t item;
    fkey_t item_key;
    fol_t item_comp;
};

Bool treat_transition( tabobj_t, long, fol_t );

static Bool
closed_completion( tabobj_t obj, struct completion_closure *argv )
{
    R_LOAD_ITEM( argv->item, argv->item_key, argv->item_comp);
    return treat_transition( obj, argv->id, argv->type );
}

Bool                            /* DYAM INSTRUCTION Dyam_Complete */
Dyam_Complete(int trans_comp_id, int item_comp_id )
{
    tabseed_t seed = R_TRANS->seed;
    void **comptab = ((void **)&seed->application_code+1);
    item_comp_t comp = ((item_comp_t) comptab)+item_comp_id;
    fol_t model = comp->model;
    fol_t item_comp = comp->item_comp;
    fol_t type = seed->type;
    struct completion_closure *closure =
        (struct completion_closure *)
        PUSH_TRAIL_BLOCK(sizeof(struct completion_closure));
    *closure = (struct completion_closure) { (void *) &closed_completion,
                                             trans_comp_id,
                                             type,
                                             R_TRANS,
                                             R_TRANS_KEY,
                                             item_comp
    };
    
    V_LEVEL_DISPLAY( V_DYAM,"  [%&e] Complete %&s %d\n",model,R_TRANS_KEY,trans_comp_id);
    return unif_retrieve_alt( model, R_TRANS_KEY, closure);
}

				/* Subsumption */

struct subsumption_closure {
    void   *fn;
    fol_t   A;
    fkey_t  Sk(A);
};

static Bool subs_result;

Bool treat_generalizer( tabobj_t, fol_t, fkey_t );

Bool
closed_subsumption( tabobj_t obj, struct subsumption_closure *argv )
{
    V_LEVEL_DISPLAY( V_DYAM, "Calling treat_generalizer %&f wrt %&f\n", TABOBJ_SUBS(obj), argv->A );
    if (treat_generalizer( obj, argv->A,argv->Sk(A) )) {
        subs_result=1;
        V_LEVEL_DISPLAY( V_DYAM, "\tFast exit!\n");
        LVALUE_R_B= REG_VALUE(R_BC);
        LVALUE_R_BC=REG_VALUE(R_BC->bc);
        V_LEVEL_DISPLAY( V_DYAM, "---------------------\n");
    }
    Succeed;
}

Bool                            /* DYAM INSTRUCTION Dyam_Subsume */
Dyam_Subsume( tabseed_t seed )
{
    fol_t A=seed->subs_comp;
    fkey_t Sk(A)=R_TRANS_KEY;
    V_LEVEL_DISPLAY( V_DYAM,"  [%&e] Subsumption %&s\n",A,Sk(A));
    if (dyalog_table) {
        struct subsumption_closure *closure=
            (struct subsumption_closure *)
            PUSH_TRAIL_BLOCK(sizeof(struct subsumption_closure));
        *closure = (struct subsumption_closure)
            { (void *)&closed_subsumption,
                  A,
                  Sk(A)
                  };
        V_LEVEL_DISPLAY( V_INDEX,"SUBS RETRIEVE %&s\n",S(A));
        subs_result=0;
        Retrieval_To_DyALog_With_Cut((fun_t)Subs_Retrieve2, S(A), closure );
        V_LEVEL_DISPLAY( V_INDEX,"Returned from Subs\n");
        return subs_result;
    } else
        Fail;
}

/**********************************************************************
 * Delete
 *      Delete every object unifiable with A_kA
 **********************************************************************/

Bool unif_delete2(TableEntry,TableEntry);
Bool unif_delete3(TableEntry, Sproto, TableEntry);
static Bool blind_delete3( TableEntry, TableEntry );
static Bool blind_delete( TableEntry );

extern void Blind_Delete2();
extern void Unif_Delete2();
extern void Unif_Delete3();

struct delete_closure {
  fol_t   A;
  fkey_t  k;
};

static void
delete_apply( cell_t *list, const Closure closure )
{
    Sdecl(A);

    A = ((struct delete_closure *) closure)->A;
    Sk(A) = ((struct delete_closure *) closure)->k;
    V_LEVEL_DISPLAY( V_INDEX, "entering delete apply %&s\n",A,Sk(A));
    for(;*list;){
        tabobj_t o = (*list)->object;
        fkey_t k = Trail_Object( o );
        fol_t  t = TABOBJ_MODEL( o );
        V_LEVEL_DISPLAY( V_INDEX, "entering delete apply trying\n",t,k);
        if (sfol_unify( A, Sk(A), t, k )) {
            *list = (*list)->next;
        } else
            list = &(*list)->next;
        Untrail_Object;
    }
}

static void
clean_entry(TableEntry parent, TableEntry entry)
{
    V_LEVEL_DISPLAY( V_INDEX,"cleaning %x %&f %x %x\n",parent,entry ? entry->key
                     : FOLNIL,entry, entry && entry->list);
    if (parent && !entry->list){
        Hash_Delete((char *)parent->list,(long)entry->key);
        if (!Hash_Nb_Elements((char *)parent->list))
            parent->list=0;
    }
}

static Bool
blind_delete3(TableEntry entry,TableEntry parent)
{
    if (entry->kind == LIST){
        delete_apply((cell_t *) &(entry->list),IP->closure);
        clean_entry(parent,entry);
    } else if (entry->son_depth > IP->depth) {
        blind_delete(entry);
        clean_entry(parent,entry);
    } else {
        V_LEVEL_DISPLAY( V_INDEX, "Restarting from blind_retrieve at depth %d\n", IP->depth);
        unif_delete2(entry,parent);
    }
    Succeed;
}

Bool
blind_delete2(TableEntry entry, HashScan *scan, TableEntry parent)
        // **DyALog Builtin** Blind_Delete2 :- blind_delete2(0,1,2)
{
    TableEntry next = (TableEntry)Hash_Next(scan);
    V_LEVEL_DISPLAY( V_INDEX,"in blind_delete2 entry key %&f\n",entry->key);
    if (next) {
        V_LEVEL_DISPLAY( V_INDEX,"Setting choice point in blind_delete2 %&f\n",next->key);
        LVALUE_X(0)=REG_VALUE(next);
        LVALUE_X(1)=REG_VALUE(scan);
        LVALUE_X(2)=REG_VALUE(parent);
        update_choice(Blind_Delete2,3);
    } else {
        Dyam_Remove_Choice();
        Delete_Topmost_Trailed_Bloc();
    }        
    return blind_delete3(entry,parent);
}

Bool
blind_delete2_start(TableEntry entry, HashScan *scan,TableEntry parent)
{
    TableEntry next = (TableEntry)Hash_Next(scan);
    V_LEVEL_DISPLAY( V_INDEX,"in blind_delete2_start entry key %&f\n",entry->key);
    if (next) {
        V_LEVEL_DISPLAY( V_INDEX,"Setting choice point in blind_delete2 %&f\n",next->key);
        LVALUE_X(0)=REG_VALUE(next);
        LVALUE_X(1)=REG_VALUE(scan);
        LVALUE_X(2)=REG_VALUE(parent);
        TRAIL_CHOICE(Blind_Delete2,3);
    } else {
        Delete_Topmost_Trailed_Bloc();
    }        
    return blind_delete3(entry,parent);
}

static Bool
blind_delete( TableEntry parent)
{
    char *table= (char *)parent->list;
    HashScan *scan=Start_HashScan();
    TableEntry entry= (TableEntry) Hash_First(table,scan);
    
    return entry && blind_delete2_start(entry,scan,parent);
}

Bool
unif_delete2(TableEntry entry,TableEntry parent)
        // **DyALog Builtin** Unif_Delete2 :- unif_delete2(0,1)
{
    V_LEVEL_DISPLAY( V_INDEX, "Unif delete\n" );
        // assert( !entry || entry->depth == depth );
    
    if (entry->kind == LIST) {
        V_LEVEL_DISPLAY( V_INDEX, "\tentry info key=%&f, depth=%d, kind=%d\n", entry->key , entry->depth, entry->kind );
        delete_apply((cell_t *) &(entry->list),IP->closure);
        clean_entry(parent,entry);
        Succeed;
    } else if ( IP->depth >= 0) {
        int depth=IP->depth;
        fol_t A = *(IP->info[depth].arg);
        fkey_t k_A = IP->info[depth].k;
        char *table = (char *) entry->list;
        fol_t  key ;
        TableEntry e1,e2;
        
        Deref(A);
        
        key  = FOLCMPP(A) ? FOLCMP_FUNCTOR(A) : A;
        
        V_LEVEL_DISPLAY( V_INDEX,"\ttiti\n");
        V_LEVEL_DISPLAY( V_INDEX, "\tcurrent term is %&f %d %&f\n", A, (long) entry,entry->key );
        V_LEVEL_DISPLAY( V_INDEX, "\tentry info key=%&f, depth=%d, kind=%d\n", entry->key , entry->depth, entry->kind );
        
        
        if (HOLE_P(A)) {                   /* try recursively all entries of table */
            V_LEVEL_DISPLAY( V_INDEX, "\tblind_delete\n");
            resume_forward();
            blind_delete(entry);
            clean_entry(parent,entry);
            Succeed;
        }
        
            /* Try variable entry in table */
        e1 = (TableEntry) Hash_Find(table,(long)VarKey);
        e2 = (TableEntry) Hash_Find(table,(long) key);
        
        if (e1) {
            if (e2) {
                V_LEVEL_DISPLAY( V_INDEX,"Setting choice point in unif_retrieve2 %&f\n",e2->key);
                LVALUE_X(0) = REG_VALUE(e2);
                LVALUE_X(1) = REG_VALUE(A);
                LVALUE_X(2) = REG_VALUE(Sk(A));
                LVALUE_X(3)= REG_VALUE(entry);
                Dyam_Full_Choice(Unif_Delete3,4);
            }
            resume_forward();
            unif_delete2(e1,entry);
            clean_entry(parent,entry);
            Succeed;
        } else if (e2) {
            resume_push(S(A));
            unif_delete2(e2,entry);
            clean_entry(parent,entry);
            Succeed;
        } else {
            Fail;
        }
    } else
        Fail;
        
}

Bool
unif_delete3(TableEntry entry, SP(A),TableEntry parent)
// **DyALog Builtin** Unif_Delete3 :- unif_delete3(0,1,2,3)
{
    Dyam_Remove_Choice();
    resume_push(S(A));
    unif_delete2(entry,parent);
    Succeed;
}

void
unif_delete( SP(A) )
{
    V_LEVEL_DISPLAY( V_INDEX, "UNIF DELETE %&s\n",S(A));
    if (dyalog_table) {
        struct delete_closure tmp = { A, Sk(A) };
        Dyam_Full_Choice(DyALog_To_C,1);
        LVALUE_X(0)= REG_VALUE(dyalog_table);
        LVALUE_X(1)= REG_VALUE(0);
        Start_Retrieval(S(A),(Closure)&tmp);
        LVALUE_R_CP= REG_VALUE((fun_t) Adjust_Address(DyALog_Fail));
//    R_BC=R_B;
        LVALUE_R_P= REG_VALUE((fun_t) Adjust_Address((fun_t)Unif_Delete2));
        C_To_DyALog();
//        Retrieval_To_DyALog((fun_t)Unif_Delete2,S(A),(Closure)&tmp);
    }
}

Bool
object_delete( tabobj_t o )
{
    cell_t *cell_handle = (cell_t *) (o->delete_info);
    V_LEVEL_DISPLAY( V_DYAM, "OBJ DELETE %x\n", (long) o);
    if (!cell_handle || (*cell_handle)->object != o){
        dyalog_error_printf("*WARNING*: trying deleting incorrect object at address %x\n",o);
        Fail;
    }
    if ((*cell_handle)->next) {
        (*cell_handle)->next->object->delete_info=(void *)cell_handle;
    }
    o->delete_info=0;
    *cell_handle=(*cell_handle)->next;
    Succeed;
}

Bool
Abolish(sfol_t symbol, const int arity) 
{
    SFOL_Deref(symbol);
    if (!FOLSMBP(symbol->t))
        Fail;
    Hash_Delete((char *)dyalog_table->list,
                (long)FOLSMB_CONVERT_ARITY(symbol->t,arity));
    Succeed;
}

